﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2015 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using System.Windows.Forms;
using HeuristicLab.Analysis;
using HeuristicLab.Collections;
using HeuristicLab.Common;
using HeuristicLab.Core.Views;
using HeuristicLab.MainForm;

namespace HeuristicLab.Optimization.Views {
  [View("ECDF Performance Comparison")]
  [Content(typeof(RunCollection), false)]
  public partial class RunCollectionECDFView : ItemView {
    private const string AllRuns = "All Runs";

    public new RunCollection Content {
      get { return (RunCollection)base.Content; }
      set { base.Content = value; }
    }

    private bool suppressUpdates;
    private readonly IndexedDataTable<double> combinedDataTable;
    public IndexedDataTable<double> CombinedDataTable {
      get { return combinedDataTable; }
    }

    public RunCollectionECDFView() {
      InitializeComponent();
      combinedDataTable = new IndexedDataTable<double>("Combined DataTable", "A data table containing the ECDF of each of a number of groups.") {
        VisualProperties = { YAxisTitle = "Proportion of reached targets" }
      };
      viewHost.Content = combinedDataTable;
      suppressUpdates = false;
    }

    #region Content events
    protected override void RegisterContentEvents() {
      base.RegisterContentEvents();
      Content.ItemsAdded += Content_ItemsAdded;
      Content.ItemsRemoved += Content_ItemsRemoved;
      Content.CollectionReset += Content_CollectionReset;
      Content.UpdateOfRunsInProgressChanged += Content_UpdateOfRunsInProgressChanged;
      Content.OptimizerNameChanged += Content_AlgorithmNameChanged;
    }
    protected override void DeregisterContentEvents() {
      Content.ItemsAdded -= Content_ItemsAdded;
      Content.ItemsRemoved -= Content_ItemsRemoved;
      Content.CollectionReset -= Content_CollectionReset;
      Content.UpdateOfRunsInProgressChanged -= Content_UpdateOfRunsInProgressChanged;
      Content.OptimizerNameChanged -= Content_AlgorithmNameChanged;
      base.DeregisterContentEvents();
    }

    private void Content_ItemsAdded(object sender, CollectionItemsChangedEventArgs<IRun> e) {
      if (suppressUpdates) return;
      if (InvokeRequired) {
        Invoke(new CollectionItemsChangedEventHandler<IRun>(Content_ItemsAdded), sender, e);
        return;
      }
      UpdateGroupComboBox();
      UpdateDataTableComboBox();
      foreach (var run in e.Items)
        RegisterRunEvents(run);
    }
    private void Content_ItemsRemoved(object sender, CollectionItemsChangedEventArgs<IRun> e) {
      if (suppressUpdates) return;
      if (InvokeRequired) {
        Invoke(new CollectionItemsChangedEventHandler<IRun>(Content_ItemsRemoved), sender, e);
        return;
      }
      UpdateGroupComboBox();
      UpdateDataTableComboBox();
      foreach (var run in e.Items)
        DeregisterRunEvents(run);
    }
    private void Content_CollectionReset(object sender, CollectionItemsChangedEventArgs<IRun> e) {
      if (suppressUpdates) return;
      if (InvokeRequired) {
        Invoke(new CollectionItemsChangedEventHandler<IRun>(Content_CollectionReset), sender, e);
        return;
      }
      UpdateGroupComboBox();
      UpdateDataTableComboBox();
      foreach (var run in e.OldItems)
        DeregisterRunEvents(run);
    }
    private void Content_AlgorithmNameChanged(object sender, EventArgs e) {
      if (InvokeRequired)
        Invoke(new EventHandler(Content_AlgorithmNameChanged), sender, e);
      else UpdateCaption();
    }
    private void Content_UpdateOfRunsInProgressChanged(object sender, EventArgs e) {
      if (InvokeRequired) {
        Invoke(new EventHandler(Content_UpdateOfRunsInProgressChanged), sender, e);
        return;
      }
      suppressUpdates = Content.UpdateOfRunsInProgress;
      if (!suppressUpdates) {
        UpdateDataTableComboBox();
        UpdateGroupComboBox();
        UpdateRuns();
      }
    }

    private void RegisterRunEvents(IRun run) {
      run.PropertyChanged += run_PropertyChanged;
    }
    private void DeregisterRunEvents(IRun run) {
      run.PropertyChanged -= run_PropertyChanged;
    }
    private void run_PropertyChanged(object sender, PropertyChangedEventArgs e) {
      if (suppressUpdates) return;
      if (InvokeRequired) {
        Invoke((Action<object, PropertyChangedEventArgs>)run_PropertyChanged, sender, e);
      } else {
        if (e.PropertyName == "Visible")
          UpdateRuns();
      }
    }
    #endregion

    protected override void OnContentChanged() {
      base.OnContentChanged();
      dataTableComboBox.Items.Clear();
      groupComboBox.Items.Clear();
      combinedDataTable.Rows.Clear();

      UpdateCaption();
      if (Content != null) {
        UpdateGroupComboBox();
        UpdateDataTableComboBox();
      }
    }

    private void UpdateRuns() {
      if (InvokeRequired) {
        Invoke((Action)UpdateRuns);
        return;
      }
      SuspendRepaint();
      try {
        combinedDataTable.VisualProperties.XAxisLogScale = false;
        combinedDataTable.Rows.Clear();
        var table = (string)dataTableComboBox.SelectedItem;
        if (string.IsNullOrEmpty(table)) return;
        var maximum = Content.Select(x => ((IndexedDataTable<double>)x.Results[table]).Rows.First().Values.Max(y => y.Item2)).Max();
        var minimum = Content.Select(x => ((IndexedDataTable<double>)x.Results[table]).Rows.First().Values.Min(y => y.Item2)).Min();
        var levels = Enumerable.Range(0, 51).Select(x => maximum - (x / 50.0) * (maximum - minimum)).ToArray();
        var selectedGroup = (string)groupComboBox.SelectedItem;
        if (string.IsNullOrEmpty(selectedGroup)) return;
        List<Tuple<string, List<IRun>>> groupedRuns;
        if (selectedGroup == AllRuns)
          groupedRuns = new List<Tuple<string, List<IRun>>> { Tuple.Create(AllRuns, Content.ToList()) };
        else groupedRuns = (from r in Content
                            group r by r.Parameters[selectedGroup].ToString() into g
                            select Tuple.Create(g.Key, g.ToList())).ToList();
        var xAxisTitles = new HashSet<string>();
        foreach (var group in groupedRuns) {
          var hits = new SortedList<double, int>();
          foreach (var run in group.Item2) {
            var resultsTable = (IndexedDataTable<double>)run.Results[table];
            xAxisTitles.Add(resultsTable.VisualProperties.XAxisTitle);
            var graph = new LinkedList<Tuple<double, double>>(resultsTable.Rows.First().Values);
            var current = graph.First.Next;
            // prune convergence graph to obtain first hit times only
            while (current != null) {
              if (current.Value.Item2.IsAlmost(current.Previous.Value.Item2)) {
                var h = current;
                current = current.Previous;
                graph.Remove(h);
              }
              current = current.Next;
            }
            var i = 0;
            current = graph.First;
            while (i < levels.Length && current != null) {
              if (current.Value.Item2 < levels[i]) {
                if (hits.ContainsKey(current.Value.Item1))
                  hits[current.Value.Item1]++;
                else hits[current.Value.Item1] = 1;
                i++;
              } else {
                current = current.Next;
              }
            }
          }
          var row = new IndexedDataRow<double>(group.Item1) { VisualProperties = { ChartType = DataRowVisualProperties.DataRowChartType.StepLine } };
          var total = 0.0;
          foreach (var h in hits) {
            total += h.Value;
            row.Values.Add(Tuple.Create(h.Key, total / (group.Item2.Count * levels.Length)));
          }
          combinedDataTable.Rows.Add(row);
        }
        combinedDataTable.VisualProperties.XAxisTitle = string.Join(" / ", xAxisTitles);
        combinedDataTable.VisualProperties.XAxisLogScale = logScalingCheckBox.Checked;
      } finally { ResumeRepaint(true); }
    }

    private void UpdateGroupComboBox() {
      string selectedItem = (string)groupComboBox.SelectedItem;

      var groupings = Content.ParameterNames.OrderBy(x => x).ToArray();
      groupComboBox.Items.Clear();
      groupComboBox.Items.Add(AllRuns);
      groupComboBox.Items.AddRange(groupings);
      if (selectedItem != null && groupComboBox.Items.Contains(selectedItem)) {
        groupComboBox.SelectedItem = selectedItem;
      } else if (groupComboBox.Items.Count > 0) {
        groupComboBox.SelectedItem = groupComboBox.Items[0];
      }
    }

    private void UpdateDataTableComboBox() {
      string selectedItem = (string)dataTableComboBox.SelectedItem;

      dataTableComboBox.Items.Clear();
      var dataTables = (from run in Content
                        from result in run.Results
                        where result.Value is IndexedDataTable<double>
                        select result.Key).Distinct().ToArray();

      dataTableComboBox.Items.AddRange(dataTables);
      if (selectedItem != null && dataTableComboBox.Items.Contains(selectedItem)) {
        dataTableComboBox.SelectedItem = selectedItem;
      } else if (dataTableComboBox.Items.Count > 0) {
        dataTableComboBox.SelectedItem = dataTableComboBox.Items[0];
      }
    }

    private void UpdateCaption() {
      Caption = Content != null ? Content.OptimizerName + " ECDF Comparison" : ViewAttribute.GetViewName(GetType());
    }

    private void groupComboBox_SelectedIndexChanged(object sender, EventArgs e) {
      UpdateRuns();
    }
    private void dataTableComboBox_SelectedIndexChanged(object sender, EventArgs e) {
      UpdateRuns();
    }

    private void logScalingCheckBox_CheckedChanged(object sender, EventArgs e) {
      combinedDataTable.VisualProperties.XAxisLogScale = logScalingCheckBox.Checked;
    }
  }
}
